import argparse
import tempfile
import importlib
from importlib.machinery import SourceFileLoader
import os
import sys
import json
from furniture.animation import Animation

try:
    from fontTools.misc.cliTools import makeOutputFileName
except:
    print("fontTools not installed")
    pass

def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return arg


def main():
    parser = argparse.ArgumentParser(
        prog="furniture-renderer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument("action", type=str)
    parser.add_argument("file", type=lambda x: is_valid_file(parser, x), metavar="FILE")
    parser.add_argument("-s", "--slice", type=str, default="")
    parser.add_argument("-f", "--folder", type=str, default=None)
    parser.add_argument("-l", "--layer", type=str, default=None)
    parser.add_argument("-v", "--verbose", type=str2bool, default=True)
    parser.add_argument("-a", "--audio", type=str2bool, default=True)
    parser.add_argument("-c", "--compile", type=str, default=None)
    parser.add_argument("-so", "--stdout", type=str, default=None)
    args = parser.parse_args()

    sl = slice(*map(lambda x: int(x.strip()) if x.strip() else None, args.slice.split(':')))

    src_prefix = os.path.basename(args.file).replace(".py", "")
    src_dirname = os.path.realpath(os.path.dirname(args.file))
    src_path = os.path.realpath(args.file)
    folder = args.folder
    if folder == None:
        folder = f"{src_dirname}/{src_prefix}_frames"
        if not os.path.exists(folder):
            os.mkdir(folder)
    else:
        folder = os.path.realpath(folder)
        if not os.path.exists(folder):
            os.mkdir(folder)

    with open(args.file, "r", encoding="utf-8") as f:
        lines = f.readlines()
    
    # act like the app
    lines.insert(0, "from drawBot import *\n")
    lines.insert(1, "__file__ = '{}'\n".format(src_path))

    if args.stdout:
        logpath = os.path.realpath(args.stdout)
        print(logpath)
        if not os.path.exists(logpath):
            with open(logpath, "w") as f:
                pass
        sys.stdout = open(logpath, "a")
        sys.stderr = open(logpath, "a")
    
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", encoding="utf-8") as temp:
        temp.write("".join(lines))
        temp.flush()
        mod = os.path.basename(temp.name)
        dirname = os.path.dirname(temp.name)
        src = SourceFileLoader(mod, temp.name).load_module()
        animation = None
        for k, obj in src.__dict__.items():
            if isinstance(obj, Animation):
                animation = obj
        
        if not animation:
            raise Exception("No furniture.animation.Animation object found in src file")

        for layer in animation.layers:
            if animation.fmt != "ufo":
                subfolder = folder + "/" + layer
                if not os.path.exists(subfolder):
                    os.mkdir(subfolder)

        if args.action == "render":
            try:
                if animation.fmt == "ufo":
                    ufo_folder = folder + "/ufos"
                    if not os.path.exists(ufo_folder):
                        os.mkdir(ufo_folder)

                animation.render(indicesSlice=sl, folder=folder, log=args.verbose)
                if args.audio:
                    os.system("afplay /System/Library/Sounds/Pop.aiff ")
                
                if args.compile:
                    if animation.fmt == "ufo":
                        ttf_folder = folder + "/ttfs"
                        if not os.path.exists(ttf_folder):
                            os.mkdir(ttf_folder)
                        # needs to support layers
                        for layer in animation.layers:
                            ttf_name = makeOutputFileName(ttf_folder + "/" + animation.name + "_" + layer, extension=".ttf")
                            ufo_path = ufo_folder + "/" + animation.name + "_" + layer + ".ufo"
                            fontmake_cmd = f"fontmake --output-path {ttf_name} -o ttf -u {ufo_path}"
                            print(fontmake_cmd)
                            os.system(fontmake_cmd)
                        os.system("afplay /System/Library/Sounds/Bottle.aiff ")
                    else:
                        print("Compliation not supported")
                        os.system("afplay /System/Library/Sounds/Basso.aiff ")
            except Exception as e:
                import traceback
                traceback.print_exc()
                if args.audio:
                    os.system("afplay /System/Library/Sounds/Sosumi.aiff ")
        
        if args.action == "info":
            print("-----")
            info = dict(animation.__dict__)
            del info["fn"]
            print(json.dumps(info))

if __name__ == "__main__":
    import sys
    sys.exit(main())


import sys
import time
import os
import json
import tempfile
import datetime
import drawBot as db
import drawBot.context.baseContext
from furniture.geometry import Rect, Edge

try:
    import defcon
except:
    pass

class RichBezier():
    def __init__(self):
        self.fill = (0, 0, 0, 1)
        self.bp = db.BezierPath()


class AnimationFrame():
    def __init__(self, animation, i):
        self.animation = animation
        self.i = i
        self.doneness = self.i / self.animation.length
        self.time = self.i / self.animation.fps
        self.data = None
        self.layers = None
        self.bps = {}

    def __repr__(self):
        return "<furniture.AnimationFrame {:04d}, {:04.2f}s, {:06.4f}%>".format(self.i, self.time, self.doneness)

    def draw(self, saving=False, saveTo=None, fmt="pdf", layers=[], fill=None):
        savingToFont = isinstance(fmt, defcon.Font)
        if saving:
            db.newDrawing()
            self.saving = True
        else:
            self.saving = False

        db.newPage(*self.animation.dimensions)
        self.page = Rect.page()

        self.bps = {}
        for l in layers:
            self.bps[l] = RichBezier()

        with db.savedState():
            if fill and not saveTo:
                with db.savedState():
                    db.fill(*fill)
                    db.rect(*self.page)
            self.layers = layers
            self.animation.fn(self)
            self.layers = None
        if self.animation.burn:
            box = self.page.take(64, Edge.MinY).take(
                120, Edge.MaxX).offset(-24, 24)
            db.fontSize(20)
            db.lineHeight(20)
            db.font("Menlo-Bold")
            db.fill(0, 0.8)
            db.rect(*box.inset(-14, -14).offset(0, 2))
            db.fill(1)
            db.textBox("{:07.2f}\n{:04d}\n{:%H:%M:%S}".format(
                self.time, self.i, datetime.datetime.now()), box, align="center")

        for k, bez in self.bps.items():
            with db.savedState():
                db.fill(*bez.fill)
                db.drawPath(bez.bp)

        if saving:
            if savingToFont:
                for k, bez in self.bps.items():
                    g = defcon.Glyph()
                    g.name = "frame_" + str(self.i)
                    g.unicode = self.i + 48 # to get to 0
                    g.width = self.animation.dimensions[0]
                    bez.bp.drawToPen(g.getPen())
                    fmt.insertGlyph(g)
            else:
                db.saveImage(f"{saveTo}/{self.i}.{fmt}")
            db.endDrawing()

        self.saving = False


class Animation():
    def __init__(self, fn,
            length=10,
            fps=30,
            dimensions=(1920, 1080),
            burn=False,
            audio=None,
            folder="frames",
            fmt="pdf",
            data=None,
            layers=["default"],
            fill=None,
            name="Animation"):
        """
        - `fn` is a callback function that takes a single argument, `frame`
        - `fps` is frames-per-second
        - `dimensions` is the page size, a tuple `(x, y)`
        - `burn`=True adds a small counter for the current frame and time
        - `audio` is not currently used
        - `folder` is the folder to which frames are rendered
        - `fmt` is file type that will be used when rendering
        - `data` is data you want sent into the callback via `frame.data`
        """
        self.fn = fn
        self.length = length
        self.fps = fps
        self.dimensions = dimensions
        self.burn = burn
        self.fmt = fmt
        self.folder = folder
        self.audio = audio
        self.data = data
        self.layers = layers
        self.fill = fill
        self.name = name

    def storyboard(self, *frames, **kwargs):
        if "frames" in kwargs:
            frames = kwargs["frames"]
        for i in frames:
            frame = AnimationFrame(self, i)
            print("(storyboard)", frame)
            #frame.data = data
            frame.draw(saving=False, saveTo=None, layers=self.layers, fill=self.fill)

    def render(self, indicesSlice=None, start=0, end=None, data=None, folder=None, fmt=None, log=True):
        if not data and self.data:
            try:
                with open(self.data, "r") as f:
                    data = json.loads(f.read())
            except FileNotFoundError:
                data = {}
        if end == None:
            end = self.length
        indices = list(range(start, end))
        
        if indicesSlice:
            indices = list(range(*indicesSlice.indices(self.length)))

        folder = folder if folder else self.folder
        fmt = fmt if fmt else self.fmt
        ufo_folder = folder + "/ufos"
        saving_to_font = fmt == "ufo"
        
        for layer in self.layers:
            # a ufo for this layer
            if saving_to_font:
                ufo_path = ufo_folder + "/" + self.name + "_" + layer + ".ufo"
                try:
                    fmt = defcon.Font(ufo_path)
                except:
                    fmt = defcon.Font()
                    fmt.save(ufo_path)
                fmt.info.familyName = self.name
                fmt.info.styleName = layer
                fmt.info.versionMajor = 1
                fmt.info.versionMinor = 0
                fmt.info.descender = 0
                fmt.info.unitsPerEm = 1000 # or self.dimensions[0] ?
                fmt.info.capHeight = self.dimensions[1]
                fmt.info.ascender = self.dimensions[1]
                fmt.info.xHeight = int(self.dimensions[1] / 2)
                fmt.save()
            
            for i in indices:
                frame = AnimationFrame(self, i)
                frame.data = data
                print(f"(render:layer:{layer})", frame)
                _folder = folder + "/" + layer
                frame.draw(saving=True, saveTo=_folder, fmt=fmt, layers=[layer], fill=self.fill)
            
            if isinstance(fmt, defcon.Font):
                fmt.save()


if __name__ == "__main__":
    def draw(frame):
        pass
    
    animation = Animation(draw, length=10, fps=30, dimensions=(1000, 1000), fmt="ufo", layers=["fg"])
    animation.storyboard(0)


from enum import Enum
import math
try:
    import drawBot as db
except ImportError:
    db = None

YOYO = "ma"

MINYISMAXY = False


class Edge(Enum):
    MaxY = 1
    MaxX = 2
    MinY = 3
    MinX = 4
    CenterY = 5
    CenterX = 6


def txt_to_edge(txt):
    if isinstance(txt, str):
        txt = txt.lower()
        if txt == "maxy":
            return Edge.MaxY
        elif txt == "maxx":
            return Edge.MaxX
        elif txt == "miny":
            return Edge.MinY
        elif txt == "minx":
            return Edge.MinX
        elif txt == "centery":
            return Edge.CenterY
        elif txt == "centerx":
            return Edge.CenterX
        else:
            return None
    else:
        return txt


def centered_square(rect):
    x, y, w, h = rect
    if w > h:
        return [x + (w - h) / 2, y, h, h]
    else:
        return [x, y + (h - w) / 2, w, w]


def perc_to_pix(rect, amount, edge):
    """
    perc(entage) to pix(els) — where the percentage is a decimal between 0 and 1 — cannot be 0 or 1
    """
    x, y, w, h = rect
    if amount < 1.0:
        d = h if edge == Edge.MinY or edge == Edge.MaxY or edge == Edge.CenterY else w
        if amount < 0:
            return d + amount
        else:
            return d * amount  # math.floor
    else:
        return amount


def divide(rect, amount, edge):
    """
    ## Dividing
    Derived from the behavior of the classic CGRectDivide, which takes a rectangle
    and breaks it into two pieces, based on a pixel amount and an edge. So if you
    want to break a rectangle 300 pixels wide into a left-hand rectangle 100 pixels
    wide and a right-hand rectangle 200 pixels wide, you could either say:
    `left, right = divide([0, 0, 300, 100], 100, Edge.MinX)`
    _or_
    `right, left = divide([0, 0, 300, 100], 200, Edge.MaxX)`
    where MaxX is the rightmost edge, and MinX is the leftmost edge

    ## Centering
    A special use-case is if you want to break a rectangle into _three_ rectangles
    based on the center "edge", you can do something like this
    `left, center, right = divide([0, 0, 300, 100], 200, Center.MinX)`
    This will result in three rectangles, always left-to-right, where
    left is 50px wide, then center is 200px wide, then right is also 50px wide —
    anything not in the center will be evenly distributed between left and right,
    or top-and-bottom in the case of a Y edge
    """
    x, y, w, h = rect
    amount = perc_to_pix(rect, amount, edge)

    if edge == Edge.MaxY:
        if MINYISMAXY:
            return [x, y, w, amount], [x, y + amount, w, h - amount]
        else:
            return [x, y + h - amount, w, amount], [x, y, w, h - amount]
    elif edge == Edge.MinY:
        if MINYISMAXY:
            return [x, y + h - amount, w, amount], [x, y, w, h - amount]
        else:
            return [x, y, w, amount], [x, y + amount, w, h - amount]
    elif edge == Edge.MinX:
        return [x, y, amount, h], [x + amount, y, w - amount, h]
    elif edge == Edge.MaxX:
        return [x + w - amount, y, amount, h], [x, y, w - amount, h]
    elif edge == Edge.CenterX:
        lw = (w - amount) / 2
        return [x, y, lw, h], [x + lw, y, amount, h], [x + lw + amount, y, lw, h]
    elif edge == Edge.CenterY:
        lh = (h - amount) / 2
        return [x, y, w, lh], [x, y + lh, w, amount], [x, y + lh + amount, w, lh]


def subdivide(rect, count, edge):
    """
    Like `divide`, but here you specify the number of equal pieces you want
    (like columns or rows), and then what edge to start at, i.e.
    `a, b, c, d, e = subdivide([0, 0, 500, 100], 5, Edge.MaxX)
    will get you five 100-px wide rectangles, right-to-left
    N.B. Does not support center edges, as that makes no sense
    """
    r = rect
    subs = []
    if hasattr(count, "__iter__"):
        amounts = count
        i = len(amounts) + 1
        a = 0
        while i > 1:
            s, r = divide(r, amounts[a], edge)
            subs.append(s)
            i -= 1
            a += 1
        subs.append(r)
        return subs
    else:
        i = count
        while i > 1:
            s, r = divide(r, 1/i, edge)
            subs.append(s)
            i -= 1
        subs.append(r)
        return subs


def pieces(rect, amount, edge):
    x, y, w, h = rect
    d = w
    if edge == Edge.MaxX or edge == Edge.MaxY:
        d = h
    fit = math.floor(d / amount)
    return subdivide(rect, fit, edge)


def take(rect, amount, edge):
    """
    Like `divide`, but here it just returns the "first" rect from a divide call,
    not all the resulting pieces, i.e. you can "take" 200px from the center of a
    rectangle by doing this
    `take([0, 0, 300, 100], 200, Edge.CenterX)`
    which will result in
    `[50, 0, 200, 100]`
    """
    if edge == Edge.CenterX or edge == Edge.CenterY:
        _, r, _ = divide(rect, amount, edge)
        return r
    else:
        r, _ = divide(rect, amount, edge)
        return r


def subtract(rect, amount, edge):
    """
    The opposite of `take`ing, this will remove and not return a piece of the
    given amount from the given edge.
    Let's say you have a 100px-wide square and you want to drop 10px from the
    right-hand side, you would do:
    `subtract([0, 0, 100, 100], 10, Edge.MaxX)`
    that leaves you with
    `[0, 0, 90, 100]`
    Actually maybe `drop` would be a better name for this, to fit a functional-
    programming vocabulary?
    """
    _, r = divide(rect, amount, edge)
    return r


def drop(rect, amount, edge):
    return subtract(rect, amount, edge)


def inset(rect, dx, dy):
    """
    Creates padding in the amount of dx and dy
    Also does expansion with negative values, or both at once
    """
    x, y, w, h = rect
    return [x + dx, y + dy, w - (dx * 2), h - (dy * 2)]


def offset(rect, dx, dy):
    x, y, w, h = rect
    if MINYISMAXY:
        return [x + dx, y - dy, w, h]
    else:
        return [x + dx, y + dy, w, h]


def expand(rect, amount, edge):
    return drop(inset(rect, amount, edge), amount/2, edge)


def centerpoint(rect):
    x, y, w, h = rect
    return [x + w/2, y + h/2]


def add(rect_a, rect_b):
    # TODO better/correct implementation!
    ax, ay, aw, ah = rect_a
    bx, by, bw, bh = rect_b
    return [
        min(ax, bx),
        min(ay, by),
        aw + bw,
        ah
    ]


def scale(rect, s, x_edge=Edge.CenterX, y_edge=Edge.CenterY):
    x, y, w, h = rect
    return [x / s, y / s, w / s, h / s]


def edgepoints(rect, edge):
    x, y, w, h = rect
    if edge == Edge.MaxY:
        if MINYISMAXY:
            return (x, y), (x + w, y)
        else:
            return (x, y + h), (x + w, y + h)
    elif edge == Edge.MinY:
        if MINYISMAXY:
            return (x, y + h), (x + w, y + h)
        else:
            return (x, y), (x + w, y)
    elif edge == Edge.MinX:
        return (x, y), (x, y + h)
    elif edge == Edge.MaxX:
        return (x + w, y), (x + w, y + h)
    elif edge == Edge.CenterX:
        return (x + w/2, y), (x + w/2, y + h)
    elif edge == Edge.CenterY:
        return (x, y + h/2), (x + w, y + h/2)


class Point():
    def __init__(self, point):
        try:
            x, y = point
            self.x = x
            self.y = y
        except:
            try:
                self.x = point.x
                self.y = point.y
            except:
                self.x = 0
                self.y = 0

    def from_obj(obj):
        p = Point((0, 0))
        try:
            p.x = obj.x
            p.y = obj.y
        except:
            pass
        return p

    def offset(self, dx, dy):
        return Point((self.x + dx, self.y + dy))

    def rect(self, w, h):
        return Rect((self.x-w/2, self.y-h/2, w, h))

    def xy(self):
        return self.x, self.y
    
    def flip(self, frame):
        return Point((self.x, frame.h - self.y))
    
    def flipSelf(self, frame):
        x, y = self.flip(frame)
        self.x = x
        self.y = y

    def __repr__(self):
        return "<furn-Point" + str(self.xy()) + ">"

    def __getitem__(self, key):
        return self.xy()[key]

    def __setitem__(self, key, value):
        if key == 0:
            self.x = value
        elif key == 1:
            self.y = value
        else:
            raise IndexError(
                "Invalid index for point assignment, must be 0 or 1")


class Rect():
    def FromCenter(center, w, h):
        x, y = center
        return Rect((x - w/2, y - h/2, w, h))

    def __init__(self, rect):
        x, y, w, h = rect
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def from_obj(obj, w=None, h=None):
        r = Rect((0, 0, 0, 0))
        try:
            r.x = obj.x
            r.y = obj.y
        except:
            pass
        try:
            r.w = obj.w
            r.h = obj.h
        except:
            pass
        if w:
            r.w = w
            r.x -= w/2
        if h:
            r.h = h
            r.y -= h/2
        return r

    def page():
        if db:
            return Rect((0, 0, db.width(), db.height()))
        else:
            raise ImportError(
                "DrawBot was not found, so `page` cannot be called")

    def __getitem__(self, key):
        return self.rect()[key]

    def __repr__(self):
        return "<furn-Rect" + str(self.rect()) + ">"

    def rect(self):
        return [self.x, self.y, self.w, self.h]

    def xy(self):
        return [self.x, self.y]

    def wh(self):
        return [self.w, self.h]

    def square(self):
        return Rect(centered_square(self.rect()))

    def divide(self, amount, edge):
        edge = txt_to_edge(edge)
        if edge == Edge.CenterX or edge == Edge.CenterY:
            a, b, c = divide(self.rect(), amount, edge)
            return Rect(a), Rect(b), Rect(c)
        else:
            a, b = divide(self.rect(), amount, edge)
            return Rect(a), Rect(b)

    def subdivide(self, amount, edge):
        edge = txt_to_edge(edge)
        return [Rect(x) for x in subdivide(self.rect(), amount, edge)]

    def subdivide_with_leadings(self, count, leadings, edge):
        edge = txt_to_edge(edge)
        leadings = leadings + [0]
        full = self.w if edge == Edge.MinX or edge == Edge.MaxX else self.h
        unit = (full - sum(leadings)) / count
        amounts = [val for pair in zip([unit] * count, leadings)
                   for val in pair][:-1]
        return [Rect(x) for x in subdivide(self.rect(), amounts, edge)][::2]

    def scale(self, s, x_edge=Edge.CenterX, y_edge=Edge.CenterY):
        x_edge = txt_to_edge(x_edge)
        y_edge = txt_to_edge(y_edge)
        return Rect(scale(self.rect(), s, x_edge, y_edge))

    def take(self, amount, edge):
        edge = txt_to_edge(edge)
        return Rect(take(self.rect(), amount, edge))

    def takeOpposite(self, amount, edge):
        edge = txt_to_edge(edge)
        return self.divide(amount, edge)[1]

    def subtract(self, amount, edge):
        edge = txt_to_edge(edge)
        return Rect(subtract(self.rect(), amount, edge))

    def expand(self, amount, edge):
        edge = txt_to_edge(edge)
        return Rect(expand(self.rect(), amount, edge))

    def inset(self, dx, dy=None):
        if dy == None:
            dy = dx
        return Rect(inset(self.rect(), dx, dy))

    def offset(self, dx, dy=None):
        if dy == None:
            dy = dx
        return Rect(offset(self.rect(), dx, dy))

    def __add__(self, another_rect):
        return Rect(add(self, another_rect))

    def grid(self, rows=2, columns=2):
        xs = [row.subdivide(columns, Edge.MinX)
              for row in self.subdivide(rows, Edge.MaxY)]
        return [item for sublist in xs for item in sublist]

    def pieces(self, amount, edge):
        edge = txt_to_edge(edge)
        return [Rect(x) for x in pieces(self.rect(), amount, edge)]

    def edge(self, edge):
        edge = txt_to_edge(edge)
        return edgepoints(self.rect(), edge)

    def center(self):
        return centerpoint(self.rect())
    
    def flip(self, frame):
        return Rect([self.x, frame.h - self.y - self.h, self.w, self.h])
    
    def flipSelf(self, frame):
        x, y, w, h = self.flip(frame)
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def point(self, eh, ev=Edge.MinX):
        ev = txt_to_edge(ev)
        if eh == "C":
            return self.point(Edge.CenterX, Edge.CenterY)
        elif eh == "W":
            return self.point(Edge.MinX, Edge.CenterY)
        elif eh == "NW":
            return self.point(Edge.MinX, Edge.MaxY)
        elif eh == "N":
            return self.point(Edge.CenterX, Edge.MaxY)
        elif eh == "NE":
            return self.point(Edge.MaxX, Edge.MaxY)
        elif eh == "E":
            return self.point(Edge.MaxX, Edge.CenterY)
        elif eh == "SE":
            return self.point(Edge.MaxX, Edge.MinY)
        elif eh == "S":
            return self.point(Edge.CenterX, Edge.MinY)
        elif eh == "SW":
            return self.point(Edge.MinX, Edge.MinY)
        else:
            px = self.x
            py = self.y

            if eh == Edge.MaxX:
                px = self.x + self.w
            elif eh == Edge.CenterX:
                px = self.x + self.w/2

            if ev == Edge.MaxY:
                py = self.y + self.h
            if ev == Edge.CenterY:
                py = self.y + self.h/2

            return Point((px, py))

if __name__ == "__main__":
    print(Rect([50, 50, 500, 500]).flip(Rect([0, 0, 1000, 1000])))


import setuptools

long_description = """
### Latest documentation available [at the github repo](https://github.com/stenson/furniture)
"""

setuptools.setup(
    name="furniture",
    version="0.2.0",
    author="Rob Stenson",
    author_email="rob.stenson@gmail.com",
    description="Typesetting and layout utilities for drawing and animating in DrawBot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stenson/furniture",
    packages=[
        "furniture"
    ],
    entry_points={
        'console_scripts': [
            'furniture = furniture.renderer:main'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import sys
import os
import importlib
from drawBot import savedState
import drawBot as db


def add_importable(path):
    """
    Add a path to the beginning of sys.path
    """
    sys.path.insert(0, os.path.expanduser(path))


def reimport(module):
    """
    Import a module from a string, and also reload it;
    also, return it
    """
    m = importlib.import_module(module)
    importlib.reload(m)
    return m


class PreparedStateManager():
    def __init__(self, style):
        self.style = style

    def __enter__(self):
        db.save()
        for k, v in self.style.items():
            db.fill(None)
            db.stroke(None)
            db.strokeWidth(1)
            getattr(db, k)(*v)
        return self

    def __exit__(self, type, value, traceback):
        db.restore()


def preparedState(**kwargs):
    """
    Pass drawBot presentation functions as dict entries
    for terser `with savedState` calls
    """
    return PreparedStateManager(kwargs)


if __name__ == "__main__":
    db.newDrawing()
    with preparedState(fill=(0, 1, 0)):
        db.fill(None)
        db.stroke(None)
        db.strokeWidth(1)
        db.rect(0, 0, 100, 100)
    db.saveImage("~/Desktop/preparedstate.png")
    db.endDrawing()

    add_importable("~/Type/grafutils")
    print(reimport("furniture.animation"))



def scale_to_axis(axis, value, insetMin=1, insetMax=1):
    """
    Scales a value between 0 and 1 to the range of an axis,
    where an 'axis' is just a dict with values for the keys
    minValue and maxValue (i.e. what’s returned from drawBot’s
    `listFontVariations` call, for a given axis)

    `insetMin` and `insetMax` let you specify an offset from the true
    min/max, in order to avoid a bug in the Apple font variation code;
    this defaults to 1 which is the most standard
    """
    minv = axis.get("minValue", 0)
    maxv = axis.get("maxValue", 1000)
    scaled = (value * (maxv - minv)) + minv
    return max(minv + insetMin, min(maxv - insetMax, scaled))


def scaledFontVariations(fs, **kwargs):
    for axis, value in kwargs.items():
        axis_dict = fs.listFontVariations().get(axis)
        scaled = scale_to_axis(axis_dict, value)
        fs.fontVariations(**{axis:scaled})
    return scaled